import 'package:args/args.dart';
import 'package:args/command_runner.dart';
import 'package:mason/mason.dart' hide packageVersion;
import 'package:pub_updater/pub_updater.dart';
import 'package:umbra_cli/src/cmd/cmd.dart';
import 'package:umbra_cli/src/commands/commands.dart';
import 'package:umbra_cli/src/platform.dart';
import 'package:umbra_cli/src/version.dart';

/// The package name.
const packageName = 'umbra_cli';

/// {@template umbra_command_runner}
/// A [CommandRunner] for the Umbra CLI.
/// {@endtemplate}
class UmbraCommandRunner extends CommandRunner<int> {
  /// {@macro umbra_command_runner}
  UmbraCommandRunner({
    Logger? logger,
    PubUpdater? pubUpdater,
    Cmd? cmd,
    Platform? platform,
  })  : _logger = logger ?? Logger(),
        _pubUpdater = pubUpdater ?? PubUpdater(),
        _cmd = cmd ?? Cmd(),
        _platform = platform ?? Platform(),
        super('umbra', 'Command Line Interface for Umbra') {
    argParser.addFlag(
      'version',
      negatable: false,
      help: 'Print the current version.',
    );
    addCommand(CreateCommand(logger: _logger, cmd: _cmd, platform: _platform));
    addCommand(
      GenerateCommand(logger: _logger, cmd: _cmd, platform: _platform),
    );
    addCommand(
      UpdateCommand(
        logger: _logger,
        cmd: _cmd,
        platform: _platform,
        pubUpdater: _pubUpdater,
      ),
    );
    addCommand(
      InstallDepsCommand(logger: _logger, cmd: _cmd, platform: _platform),
    );
  }

  /// Standard timeout duration for the CLI.
  static const timeout = Duration(milliseconds: 500);

  final Logger _logger;
  final PubUpdater _pubUpdater;
  final Cmd _cmd;
  final Platform _platform;

  @override
  Future<int> run(Iterable<String> args) async {
    try {
      final _argResults = parse(args);
      return await runCommand(_argResults) ?? ExitCode.success.code;
    } on FormatException catch (e, stackTrace) {
      _logger
        ..err(e.message)
        ..err('$stackTrace')
        ..info('')
        ..info(usage);
      return ExitCode.usage.code;
    } on UsageException catch (e) {
      _logger
        ..err(e.message)
        ..info('')
        ..info(e.usage);
      return ExitCode.usage.code;
    }
  }

  @override
  Future<int?> runCommand(ArgResults topLevelResults) async {
    int? exitCode = ExitCode.unavailable.code;
    if (topLevelResults['version'] == true) {
      _logger.info(packageVersion);
      exitCode = ExitCode.success.code;
    } else {
      exitCode = await super.runCommand(topLevelResults);
    }
    if (topLevelResults.command?.name != 'update') await _checkForUpdates();
    return exitCode;
  }

  Future<void> _checkForUpdates() async {
    try {
      final latestVersion = await _pubUpdater.getLatestVersion(packageName);
      final isUpToDate = packageVersion == latestVersion;
      if (!isUpToDate) {
        _logger
          ..info('')
          ..info(
            '''
${lightYellow.wrap('Update available!')} ${lightCyan.wrap(packageVersion)} \u2192 ${lightCyan.wrap(latestVersion)}
${lightYellow.wrap('Changelog:')} ${lightCyan.wrap('https://github.com/wolfenrain/umbra/releases/tag/umbra_cli-v$latestVersion')}
Run ${cyan.wrap('umbra update')} to update''',
          );
      }
    } catch (_) {}
  }
}
